# This file is deprecated. TODO(Patrick) - as soon as vllm & transformers have adapted
# to new naming, throw deprecation warning and remove this file.
from mistral_common.image import *  # noqa: F403
